(*---------------------------------------------*)
(* (c) Mike Willems 1986/5                     *)
(* LIFE - the famous Conway Cellular Automaton *)
(* Turbo-Pascal implementation                 *)
(* For EPSON PX-8 and other computers. If      *)
(* necessary, just change Constants.           *)
(*---------------------------------------------*)


Program Life;

const
  MaxX = 40;
  MaxY = 8;
  CurR = #$1C;
  CurL = #$1D;
  CurU = #$1E;
  CurD = #$1F;

type
  Cell = record
           CurrentState, NewState  : (Live,Dead);
           AdjacentCount : 0..8;
         end;

var
  Universe : Array[1..MaxX,1..MaxY] of Cell;
  x,y : Byte;
  p,q : Byte;
  Gencount : Integer;
  Chr : Char;
  Step: Boolean;


Procedure ResetUniverse;
begin
  For x:=1 to MaxX do
  begin
    For y:=1 to MaxY do
    begin
      with Universe [x,y] do
      begin
        CurrentState := Dead;
        NewState     := Dead;
        AdjacentCount:= 0;
      end;
    end;
  end;
end;


Procedure God; (*BIG, but not all He's made out to be - no free will*)

Procedure ResetCounter; (* God Cleans Things Up *)
begin
  For x:=1 to MaxX do
  begin
    For y:=1 to MaxY do
    With Universe[x,y] do
    begin
      If AdjacentCount <>0 then AdjacentCount := 0;
    end;
  end;
end;

Procedure CountNeighbours; (* God holds a Census *)
var
  r,s : Byte;
begin
  ResetCounter;
  For X:=1 to MaxX do
  begin
    For Y:=1 to MaxY do
    begin
      (* if live, all neighbors count+1 *)
      If Universe[x,y].Currentstate = Live then
      begin
        (*make all neighbors count+1*)
        For p:=x-1 to x+1 do
        begin
          For q:=y-1 to y+1 do
          begin
            (*if not outside borders or own field, then incr.count*)
            If (p<>x) OR (q<>y) then   (* not own field *)
            begin
              If p=0 then r:=MaxX else if p>MaxX then r:=1 else r:=p;
              If q=0 then s:=MaxY else if q>MaxY then s:=1 else s:=q;
              with Universe[r,s] do
              begin
                AdjacentCount := AdjacentCount+1;
              end;
            end;
          end;
        end;
      end;
    end;
  end;
end;

begin (* Now, God giveth and taketh away *)
  CountNeighbours;
  For X:=1 to MaxX do
  begin
    For Y:=1 to MaxY do
    begin
      (* HERE, THE LIFE-RULES: *)
      With Universe[x,y] do
      begin
        If CurrentState = Live then
        begin
          If (AdjacentCount<2) OR (AdjacentCount>3) then NewState:=Dead
          else NewState:=Live;
        end
        else
        begin
          If AdjacentCount=3 then NewState := Live else NewState:=Dead;
        end;
      end;
    end;
  end;
end; (* Yes, God has an end *)


Procedure PrintUniverse;
begin
  For Y:=1 to MaxY do
  begin
    For X:= 1 to MaxX do
    begin
      (* compare, print if changed *)
      With Universe[x,y] do
      begin
        If NewState<>CurrentState then
        begin
          GotoXY(X,Y);
          If NewState = Live then Write('*') else write(' ');
          CurrentState := NewState;
        end;
      end;
    end;
  end;
end;


Procedure EditUniverse;
var
  Chr : char;

Procedure ClearUniverse; (* internal to Edit *)
begin
  For X:=1 to MaxX do
  begin
    For Y:=1 to MaxY do
    begin
      With Universe[x,y] do
      If CurrentState=Live then NewState:=Dead;
    end;
  end;
end;

begin
  Step := False;
  p:=20; q:=5; GotoXY(45,7);Write('EDIT mode : Kill/Live/Run/Step/Cle');
  Repeat
    GotoXY(p,q);
    Read (Kbd, Chr);
    Case (Chr) of
      CurR : begin p:=p+1; If p>MaxX then p:=1; end; (*right*)
      CurL : begin p:=p-1; If p<1 then p:=MaxX; end; (*left*)
      CurU : begin q:=q-1; If q<1 then q:=MaxY; end; (*up*)
      CurD : begin q:=q+1; If q>MaxY then q:=1; end; (*down*)
      's' : Step:=True;
      'c' : begin ClearUniverse; GenCount:=1; PrintUniverse; end;
      'l' : begin Universe[p,q].CurrentState := Live; Write('*'); end;
      'k' : begin Universe[p,q].CurrentState := Dead; Write(' '); end;
    end;
  until Chr in ['r','s'];
  GotoXY(45,7);
  If Chr='r' then Write('RUN  mode : Edit/eXit/Step/Run    ') else
                  Write('STEP mode : Edit/eXit/Step/Run    ');
end;


(* MAIN BLOCK *)
begin
  ClrScr; GenCount :=1;
  ResetUniverse;
  GotoXY(45,8);Write('Generation: 0');
  EditUniverse;
  Repeat
    Chr:=' ';
    God;
    PrintUniverse;
    GotoXY(57,8); Write(GenCount:3);
    GenCount:=GenCount+1;
    If Step=True then repeat until keypressed;
    If KeyPressed then Read(Kbd,Chr);
    Case(Chr) of
      'e': EditUniverse;
      'r': begin Step := False; GotoXY(45,7); Write('RUN '); end;
      's': begin Step := True;  GotoXY(45,7); Write('STEP'); end;
    end;
  Until Chr='x';
end.
