program PAL;   { ----------------------------- }       {         By:         }
               {   P A S C A L   L I N K E R   }       { Jean-Claude Wippler }
               { ----------------------------- }       {    Meekrap oord 6   }
               {  For MS-DOS and CP/M systems  }       {   3991 VE  Houten   }
               { Usage:  PAL src.PAS [out.PAI] }       {      Nederland      }

const
  VERSION = 1.2;        { May 10, 1987: continuation lines }
  MAXREF  = 500;        { maximum number of references     }
  DEFLIB  = 'STDLIB';   { default name of source library   }
  SUFLIB  = '.PAL';     { default suffix of source library }
  SUFSRC  = '.PAS';     { default suffix of main source    }
  SUFINC  = '.PAI';     { suffix of include output file    }

type
  Name = string [80];                                         {! For MS-DOS !}
{ Name = string [20];                                         {!  For CP/M  !}
  StrL = string [150];

  PtrPtr = ^ SymPtr;
  SymPtr = ^ Symbol;
  Symbol = record
             nam: Name;
             nxt: SymPtr
           end;

  SrcRef = record
             src: Name;
             lin: integer;
             inc: boolean;
             def: SymPtr;
             ext: SymPtr
           end;

var
  reference: array [0..MAXREF] of ^ SrcRef;   { all references encountered }
  lastref: -1..MAXREF;                        { last reference slot used   }
  include: SrcRef;                            { output and lib file names  }
  defined: SymPtr;                            { list of defined references }
  undefined: SymPtr;                          { list of undefined refs     }

         { Operating-system dependent pointer manipulation }

function Pointer(var p): PtrPtr;
begin
  Pointer := Ptr(Seg(p), Ofs(p))                              {! For MS-DOS !}
{ Pointer := Ptr(Addr(p))                                     {!  For CP/M  !}
end;

         { Create a new reference in specified source file and line }

function NewRef(nam: Name; num: integer): integer;
begin
  lastref := lastref + 1;
  if lastref > MAXREF then
    begin
      WriteLn;
      WriteLn('Cannot process more than ', MAXREF, ' references.');
      Halt
    end;
  New(reference[lastref]);
  with reference[lastref]^ do
    begin
      src := nam;
      lin := num;
      inc := False;
      def := Nil;
      ext := Nil
    end;
  NewRef := lastref
end;

         { Check whether symbol is present on specified list }

function OnList(sym: Name; pnt: SymPtr): Boolean;
begin
  OnList := True;
  while pnt <> Nil do
    with pnt^ do
      if nam = sym then
        Exit
      else
        pnt := nxt;
  OnList := False
end;

         { Append symbol at end of list if not yet present }

procedure AddList(sym: Name; var pnt: SymPtr);
  var p: PtrPtr;
begin
  p := Pointer(pnt);
  while p^ <> Nil do
    with p^^ do
      if nam = sym then
        Exit
      else
        p := Pointer(nxt);
  New(p^);
  p^^.nam := sym;
  p^^.nxt := Nil
end;

         { Remove symbol from list if present }

procedure RemList(sym: Name; var pnt: SymPtr);
  var p: PtrPtr; q: SymPtr;
begin
  p := Pointer(pnt);
  while p^ <> Nil do
    with p^^ do
      if nam = sym then
        begin
          q := p^;
          p^ := nxt;
          Dispose(q);
          Exit
        end
      else
        p := Pointer(nxt)
end;

         { Scan string and append all symbols encountered to given list }

procedure ParseList(lin: StrL; var ptr: SymPtr);
  var i: byte;
begin
  repeat
    lin := lin + ' ?';
    while lin[1] in [' ', ','] do
      Delete(lin, 1, 1);
    i := 0;
    while lin[i+1] in ['a'..'z','A'..'Z','0'..'9','_','.',':'] do
      i := i + 1;
    if i = 0 then
      Exit;
    AddList(Copy(lin, 1, i), ptr);
    Delete(lin, 1, i)
  until False
end;

         { Scan source or library file with optional default suffix }

procedure ScanSource(nam, suf: Name);
  var i, j, n: integer; f: text; l: StrL;
begin
  if Pos('.', nam) = 0 then
    nam := nam + suf;
  Write(nam, ' ');
  Assign(f, nam);
  Reset(f);
  n := 0;
  while not Eof(f) do
    begin
      n := n + 1;
      ReadLn(f, l);
      if (Length(l) >= 3) and (l[1] = '{') and (l[2] in ['$','#']) then
        begin
          for i := 1 to Length(l) do
            l[i] := UpCase(l[i]);
          i := Pos(SUFINC, l);
          if (Copy(l, 2, 3) = '$I ') and (i > 0) then
            begin
              AddList(Copy(l, 5, i-1), include.def);
              if Copy(l, i+4, 3) = ' = ' then
                ParseList(Copy(l, i+7, 255), include.ext)
            end
          else if l[2] = '#' then
            begin
              i := Pos(':', l);
              if i <= 4 then
                j := NewRef(nam, n)
              else if Copy(l, i-2, 2) <> '  ' then
                j := NewRef(nam, n)
              else
                j := lastref;
              if i > 0 then
                begin
                  ParseList(Copy(l, i+1, 255), reference[j]^.ext);
                  Delete(l, i, 255)
                end;
              ParseList(Copy(l, 3, 255), reference[j]^.def)
            end
        end
    end;
  Close(f)
end;

         { Append all not yet defined external references to undefined list }

procedure Need(p: SymPtr);
begin
  while p <> Nil do
    with p^ do
      begin
        if not OnList(nam, defined) then
          AddList(nam, undefined);
        p := nxt
      end
end;

         { Resolve max possible with new references, given starting index }

procedure Resolve(fst: integer);
  var i: integer; done: boolean;
begin
  repeat
    done := True;
    for i := fst to lastref do
      with reference[i]^ do
        if def <> Nil then
          if OnList(def^.nam, undefined) then
            begin
              done := False;
              RemList(def^.nam, undefined);
              AddList(def^.nam, defined);
              inc := True;
              Need(ext)
            end
  until done
end;

         { Extract contents starting at given reference to output file }

procedure Extract(pos: integer; var out: text);
  var c: byte; n: integer; m: Name; l: StrL; f: text; done, xfer: boolean;
begin
  m := reference[pos]^.src;
  WriteLn(out);
  WriteLn(out, '{===< ', m, ': >===}');
  WriteLn(out);
  Assign(f, m);
  Reset(f);
  n := 0;
  xfer := false;
  c := Length(m) + 2;
  Write(m, ': ');
  repeat
    done := Eof(f);
    if not done then
      begin
        n := n + 1;
        ReadLn(f, l);
        with reference[pos]^ do
          if (src = m) and (lin = n) then
            begin
              if def = Nil then
                xfer := True
              else
                if inc then
                  with def^ do
                    begin
                      if c + Length(nam) > 78 then
                        begin WriteLn; Write('':4); c := 4 end;
                      Write(nam, ' ');
                      c := c + Length(nam) + 1;
                      RemList(def^.nam, defined);
                      xfer := True
                    end
                else
                  xfer := False;
              pos := pos + 1
            end;
        if xfer then
          WriteLn(out, l)
      end
  until done;
  Close(f);
  WriteLn
end;

         { Create output file and extract necessary definitions }

procedure MakeLib(out: Name);
  var i, j: integer; f: text;
begin
  if Pos('.', out) = 0 then
    out := out + SUFINC;
  WriteLn('Building: ', out);
  WriteLn;
  Assign(f, out);
  Rewrite(f);
  WriteLn(f, '{===< PAL output file >===}');
  i := lastref;
  repeat
    repeat
      i := i - 1
    until reference[i]^.src <> reference[i+1]^.src;
    if i > 0 then
      Extract(i+1, f)
  until i = 0;
  WriteLn(f);
  WriteLn(f, '{===< eof >===}');
  Close(f)
end;

         { Find undefined external reference }

function FindRef(s: Name): Name;
  var i: integer;
begin
  for i := 1 to lastref do
    with reference[i]^ do
      if OnList(s, ext) then
        begin
          if def = Nil then
            FindRef := src
          else
            FindRef := def^.nam;
          Exit
        end;
  FindRef := '???';
end;

         { Main pogram }

var
  i: integer;
  s: Name;

begin
  lastref := -1;
  i := NewRef('?', 0);
  include.def := Nil;
  include.ext := Nil;
  defined := Nil;
  undefined := Nil;

  if ParamCount = 0 then
    begin
      WriteLn('PAscal Linker, V', VERSION:0:1, ' by J.C.Wippler.');
      WriteLn;
      WriteLn('Usage: PAL src', SUFSRC, ' [out', SUFINC, ']');
      Halt
    end;
  s := ParamStr(1);

  Write('Scanning: ');
  ScanSource(s, SUFSRC);
  for i := 1 to lastref do
    Need(reference[i]^.ext);
  while include.ext <> Nil do
    begin
      if undefined <> Nil then
        begin
          i := lastref + 1;
          ScanSource(include.ext^.nam, SUFLIB);
          Resolve(i)
        end;
      include.ext := include.ext^.nxt
    end;
  WriteLn;
  if undefined <> Nil then
    begin
      WriteLn('Unsatisfied externals:');
      repeat
        WriteLn('   ', undefined^.nam, ' (in ', FindRef(undefined^.nam), ')');
        undefined := undefined^.nxt
      until undefined = Nil;
      WriteLn('No include file generated.')
    end
  else
    if ParamCount > 1 then
      MakeLib(ParamStr(2))
    else if include.def = Nil then
      MakeLib(DEFLIB)
    else
      MakeLib(include.def^.nam)
end.
